@@ -1,13 +1,16 @@
+{-# LANGUAGE CPP #-}
+
 module Network.Wai.Handler.Warp.Conduit where
 
 import Control.Exception
 import Control.Monad (when, unless)
 import Data.ByteString (ByteString)
-import Data.ByteString.Lazy.Char8 (pack)
 import qualified Data.ByteString as S
-import qualified Data.ByteString.Lazy as L
 import qualified Data.IORef as I
-import Data.Word (Word, Word8)
+#if __GLASGOW_HASKELL__ < 709
+import Data.Word (Word)
+#endif
+import Data.Word (Word8)
 import Network.Wai.Handler.Warp.Types
 
 ----------------------------------------------------------------
@@ -76,9 +79,6 @@ data ChunkState = NeedLen
                 | DoneChunking
     deriving Show
 
-bsCRLF :: L.ByteString
-bsCRLF = pack "\r\n"
-
 mkCSource :: Source -> IO CSource
 mkCSource src = do
     ref <- I.newIORef NeedLen
@@ -148,13 +148,13 @@ readCSource (CSource src ref) = do
                 return S.empty
             else do
                 (x, y) <-
-                    case S.breakByte 10 bs of
+                    case S.break (== 10) bs of
                         (x, y)
                             | S.null y -> do
                                 bs2 <- readSource' src
-                                if S.null bs2
-                                    then return (x, y)
-                                    else return $ S.breakByte 10 $ bs `S.append` bs2
+                                return $ if S.null bs2
+                                    then (x, y)
+                                    else S.break (== 10) $ bs `S.append` bs2
                             | otherwise -> return (x, y)
                 let w =
                         S.foldl' (\i c -> i * 16 + fromIntegral (hexToWord c)) 0
