@@ -6,6 +6,7 @@
 module Network.Wai.Handler.Warp.Request (
     recvRequest
   , headerLines
+  , pauseTimeoutKey
   ) where
 
 import qualified Control.Concurrent as Conc (yield)
@@ -16,7 +17,6 @@ import qualified Data.ByteString as S
 import qualified Data.ByteString.Unsafe as SU
 import qualified Data.CaseInsensitive as CI
 import qualified Data.IORef as I
-import Data.Monoid (mempty)
 import qualified Network.HTTP.Types as H
 import Network.Socket (SockAddr)
 import Network.Wai
@@ -30,6 +30,8 @@ import Network.Wai.Handler.Warp.Types
 import Network.Wai.Internal
 import Prelude hiding (lines)
 import Control.Monad (when)
+import qualified Data.Vault.Lazy as Vault
+import System.IO.Unsafe (unsafePerformIO)
 
 ----------------------------------------------------------------
 
@@ -47,9 +49,13 @@ recvRequest :: Settings
             -> SockAddr -- ^ Peer's address.
             -> Source -- ^ Where HTTP request comes from.
             -> IO (Request
-                  ,IndexedHeader) -- ^
+                  ,Maybe (I.IORef Int)
+                  ,IndexedHeader
+                  ,IO ByteString) -- ^
             -- 'Request' passed to 'Application',
+            -- how many bytes remain to be consumed, if known
             -- 'IndexedHeader' of HTTP request for internal use,
+            -- Body producing action used for flushing the request body
 
 recvRequest settings conn ii addr src = do
     hdrlines <- headerLines src
@@ -58,9 +64,12 @@ recvRequest settings conn ii addr src = do
         expect = idxhdr ! idxExpect
         cl = idxhdr ! idxContentLength
         te = idxhdr ! idxTransferEncoding
-    handleExpect conn httpversion expect
-    (rbody, bodyLength) <- bodyAndSource src cl te
-    rbody' <- timeoutBody th rbody
+        handle100Continue = handleExpect conn httpversion expect
+    (rbody, remainingRef, bodyLength) <- bodyAndSource src cl te
+    -- body producing function which will produce '100-continue', if needed
+    rbody' <- timeoutBody remainingRef th rbody handle100Continue
+    -- body producing function which will never produce 100-continue
+    rbodyFlush <- timeoutBody remainingRef th rbody (return ())
     let req = Request {
             requestMethod     = method
           , httpVersion       = httpversion
@@ -72,12 +81,14 @@ recvRequest settings conn ii addr src = do
           , isSecure          = False
           , remoteHost        = addr
           , requestBody       = rbody'
-          , vault             = mempty
+          , vault             = Vault.insert pauseTimeoutKey
+                                (Timeout.pause th)
+                                Vault.empty
           , requestBodyLength = bodyLength
           , requestHeaderHost = idxhdr ! idxHost
           , requestHeaderRange = idxhdr ! idxRange
           }
-    return (req, idxhdr)
+    return (req, remainingRef, idxhdr, rbodyFlush)
   where
     th = threadHandle ii
 
@@ -87,7 +98,7 @@ headerLines :: Source -> IO [ByteString]
 headerLines src = do
     bs <- readSource src
     if S.null bs
-        then throwIO $ NotEnoughLines []
+        then throwIO ConnectionClosedByPeer
         else push src (THStatus 0 id id) bs
 
 ----------------------------------------------------------------
@@ -111,15 +122,16 @@ bodyAndSource :: Source
               -> Maybe HeaderValue -- ^ content length
               -> Maybe HeaderValue -- ^ transfer-encoding
               -> IO (IO ByteString
+                    ,Maybe (I.IORef Int)
                     ,RequestBodyLength
                     )
 bodyAndSource src cl te
   | chunked = do
       csrc <- mkCSource src
-      return (readCSource csrc, ChunkedBody)
+      return (readCSource csrc, Nothing, ChunkedBody)
   | otherwise = do
-      isrc <- mkISource src len
-      return (readISource isrc, bodyLen)
+      isrc@(ISource _ remaining) <- mkISource src len
+      return (readISource isrc, Just remaining, bodyLen)
   where
     len = toLength cl
     bodyLen = KnownLength $ fromIntegral len
@@ -135,18 +147,35 @@ isChunked _         = False
 
 ----------------------------------------------------------------
 
-timeoutBody :: Timeout.Handle -> IO ByteString -> IO (IO ByteString)
-timeoutBody timeoutHandle rbody = do
+timeoutBody :: Maybe (I.IORef Int) -- ^ remaining
+            -> Timeout.Handle
+            -> IO ByteString
+            -> IO ()
+            -> IO (IO ByteString)
+timeoutBody remainingRef timeoutHandle rbody handle100Continue = do
     isFirstRef <- I.newIORef True
 
+    let checkEmpty =
+            case remainingRef of
+                Nothing -> return . S.null
+                Just ref -> \bs -> if S.null bs
+                    then return True
+                    else do
+                        x <- I.readIORef ref
+                        return $! x <= 0
+
     return $ do
         isFirst <- I.readIORef isFirstRef
 
-        when isFirst $
+        when isFirst $ do
+            -- Only check if we need to produce the 100 Continue status
+            -- when asking for the first chunk of the body
+            handle100Continue
             -- Timeout handling was paused after receiving the full request
             -- headers. Now we need to resume it to avoid a slowloris
             -- attack during request body sending.
             Timeout.resume timeoutHandle
+            I.writeIORef isFirstRef False
 
         bs <- rbody
 
@@ -154,7 +183,8 @@ timeoutBody timeoutHandle rbody = do
         -- because the application is not interested in more bytes, or
         -- because there is no more data available, pause the timeout
         -- handler again.
-        when (S.null bs) (Timeout.pause timeoutHandle)
+        isEmpty <- checkEmpty bs
+        when isEmpty (Timeout.pause timeoutHandle)
 
         return bs
 
@@ -203,7 +233,7 @@ push src (THStatus len lines prepend) bs'
     -- update the length, and continue processing.
     push' Nothing = do
         bst <- readSource' src
-        when (S.null bs) $ throwIO IncompleteHeaders
+        when (S.null bst) $ throwIO IncompleteHeaders
         push src status bst
       where
         len' = len + bsLen
@@ -241,6 +271,10 @@ push src (THStatus len lines prepend) bs'
 
 {-# INLINE checkCR #-}
 checkCR :: ByteString -> Int -> Int
-checkCR bs pos = if 13 == S.index bs p then p else pos -- 13 is CR
+checkCR bs pos = if pos > 0 && 13 == S.index bs p then p else pos -- 13 is CR
   where
     !p = pos - 1
+
+pauseTimeoutKey :: Vault.Key (IO ())
+pauseTimeoutKey = unsafePerformIO Vault.newKey
+{-# NOINLINE pauseTimeoutKey #-}
