@@ -9,43 +9,53 @@ module Network.Wai.Handler.Warp.Response (
   , fileRange -- for testing
   , warpVersion
   , defaultServerValue
+  , addDate
+  , addServer
+  , hasBody
   ) where
 
 #ifndef MIN_VERSION_base
 #define MIN_VERSION_base(x,y,z) 1
 #endif
 
-import Blaze.ByteString.Builder (fromByteString, Builder, flush)
 import Blaze.ByteString.Builder.HTTP (chunkedTransferEncoding, chunkedTransferTerminator)
+#if __GLASGOW_HASKELL__ < 709
 import Control.Applicative
+#endif
 import Control.Exception
+import Control.Monad (unless, when)
 import Data.Array ((!))
 import Data.ByteString (ByteString)
-import Data.Streaming.Blaze (newBlazeRecv, reuseBufferStrategy)
 import qualified Data.ByteString as S
-import Control.Monad (unless, when)
+import Data.ByteString.Builder (byteString, Builder)
+import Data.ByteString.Builder.Extra (flush)
 import qualified Data.ByteString.Char8 as B (pack)
 import qualified Data.CaseInsensitive as CI
 import Data.Function (on)
 import Data.List (deleteBy)
 import Data.Maybe (isJust, listToMaybe)
 #if MIN_VERSION_base(4,5,0)
-import Data.Monoid ((<>), mempty)
+# if __GLASGOW_HASKELL__ < 709
+import Data.Monoid (mempty)
+# endif
+import Data.Monoid ((<>))
 #else
 import Data.Monoid (mappend, mempty)
 #endif
+import Data.Streaming.Blaze (newBlazeRecv, reuseBufferStrategy)
 import Data.Version (showVersion)
 import qualified Network.HTTP.Types as H
 import Network.Wai
+import Network.Wai.Handler.Warp.Buffer (toBuilderBuffer)
 import qualified Network.Wai.Handler.Warp.Date as D
-import Network.Wai.Handler.Warp.Buffer (toBlazeBuffer)
+import qualified Network.Wai.Handler.Warp.FdCache as F
 import Network.Wai.Handler.Warp.Header
 import Network.Wai.Handler.Warp.IO (toBufIOWith)
-import Network.Wai.Handler.Warp.ResponseHeader
 import Network.Wai.Handler.Warp.RequestHeader (parseByteRanges)
+import Network.Wai.Handler.Warp.ResponseHeader
 import qualified Network.Wai.Handler.Warp.Timeout as T
 import Network.Wai.Handler.Warp.Types
-import Network.Wai.Internal
+import Network.Wai.Internal (Response (ResponseBuilder, ResponseStream, ResponseFile, ResponseRaw))
 import Numeric (showInt)
 import qualified Paths_warp
 import qualified System.PosixCompat.Files as P
@@ -174,9 +184,9 @@ sendResponse :: ByteString -- ^ default server value
              -> IO Bool -- ^ Returing True if the connection is persistent.
 sendResponse defServer conn ii req reqidxhdr src response = do
     hs <- addServerAndDate hs0
-    if hasBody s req then do
+    if hasBody s then do
         -- HEAD comes here even if it does not have body.
-        sendRsp conn ver s hs rsp
+        sendRsp conn mfdc ver s hs rsp
         T.tickle th
         return ret
       else do
@@ -190,10 +200,12 @@ sendResponse defServer conn ii req reqidxhdr src response = do
     rspidxhdr = indexResponseHeader hs0
     th = threadHandle ii
     dc = dateCacher ii
+    mfdc = fdCacher ii
     addServerAndDate = addDate dc rspidxhdr . addServer defServer rspidxhdr
     mRange = reqidxhdr ! idxRange
-    reqinfo@(isPersist,_) = infoFromRequest req reqidxhdr
-    (isKeepAlive, needsChunked) = infoFromResponse rspidxhdr reqinfo
+    (isPersist,isChunked0) = infoFromRequest req reqidxhdr
+    isChunked = not isHead && isChunked0
+    (isKeepAlive, needsChunked) = infoFromResponse rspidxhdr (isPersist,isChunked)
     isHead = requestMethod req == H.methodHead
     rsp = case response of
         ResponseFile _ _ path mPart -> RspFile path mPart mRange isHead (T.tickle th)
@@ -216,39 +228,52 @@ data Rsp = RspFile FilePath (Maybe FilePart) (Maybe HeaderValue) Bool (IO ())
 ----------------------------------------------------------------
 
 sendRsp :: Connection
+        -> Maybe F.MutableFdCache
         -> H.HttpVersion
         -> H.Status
         -> H.ResponseHeaders
         -> Rsp
         -> IO ()
-sendRsp conn ver s0 hs0 (RspFile path mPart mRange isHead hook) = do
+sendRsp conn mfdc ver s0 hs0 (RspFile path mPart mRange isHead hook) = do
     ex <- fileRange s0 hs path mPart mRange
     case ex of
         Left _ex ->
 #ifdef WARP_DEBUG
           print _ex >>
 #endif
-          sendRsp conn ver s2 hs2 (RspBuilder body True)
+          sendRsp conn mfdc ver s2 hs2 (RspBuilder body True)
         Right (s, hs1, beg, len)
           | len >= 0 ->
             if isHead then
-                sendRsp conn ver s hs1 (RspBuilder mempty False)
+                sendRsp conn mfdc ver s hs1 (RspBuilder mempty False)
               else do
                 lheader <- composeHeader ver s hs1
-                connSendFile conn path beg len hook [lheader]
-          | otherwise -> do
-            sendRsp conn ver H.status416
+#ifdef WINDOWS
+                let fid = FileId path Nothing
+                    hook' = hook
+#else
+                (mfd, hook') <- case mfdc of
+                   -- settingsFdCacheDuration is 0
+                   Nothing  -> return (Nothing, hook)
+                   Just fdc -> do
+                      (fd, fresher) <- F.getFd fdc path
+                      return (Just fd, hook >> fresher)
+                let fid = FileId path mfd
+#endif
+                connSendFile conn fid beg len hook' [lheader]
+          | otherwise ->
+            sendRsp conn mfdc ver H.status416
                 (filter (\(k, _) -> k /= "content-length") hs1)
                 (RspBuilder mempty True)
   where
     hs = addAcceptRanges hs0
     s2 = H.status404
     hs2 =  replaceHeader H.hContentType "text/plain; charset=utf-8" hs0
-    body = fromByteString "File not found"
+    body = byteString "File not found"
 
 ----------------------------------------------------------------
 
-sendRsp conn ver s hs (RspBuilder body needsChunked) = do
+sendRsp conn _ ver s hs (RspBuilder body needsChunked) = do
     header <- composeHeaderBuilder ver s hs needsChunked
     let hdrBdy
          | needsChunked = header <> chunkedTransferEncoding body
@@ -260,10 +285,10 @@ sendRsp conn ver s hs (RspBuilder body needsChunked) = do
 
 ----------------------------------------------------------------
 
-sendRsp conn ver s hs (RspStream streamingBody needsChunked th) = do
+sendRsp conn _ ver s hs (RspStream streamingBody needsChunked th) = do
     header <- composeHeaderBuilder ver s hs needsChunked
     (recv, finish) <- newBlazeRecv $ reuseBufferStrategy
-                    $ toBlazeBuffer (connWriteBuffer conn) (connBufferSize conn)
+                    $ toBuilderBuffer (connWriteBuffer conn) (connBufferSize conn)
     let send builder = do
             popper <- recv builder
             let loop = do
@@ -283,7 +308,7 @@ sendRsp conn ver s hs (RspStream streamingBody needsChunked th) = do
 
 ----------------------------------------------------------------
 
-sendRsp conn _ _ _ (RspRaw withApp src tickle) =
+sendRsp conn _ _ _ _ (RspRaw withApp src tickle) =
     withApp recv send
   where
     recv = do
@@ -356,13 +381,12 @@ infoFromResponse rspidxhdr (isPersist,isChunked) = (isKeepAlive, needsChunked)
 
 ----------------------------------------------------------------
 
-hasBody :: H.Status -> Request -> Bool
-hasBody s req = sc /= 204
-             && sc /= 304
-             && sc >= 200
+hasBody :: H.Status -> Bool
+hasBody s = sc /= 204
+         && sc /= 304
+         && sc >= 200
   where
     sc = H.statusCode s
-    method = requestMethod req
 
 ----------------------------------------------------------------
 
@@ -385,9 +409,9 @@ addContentRange beg end total hdrs = (hContentRange, range) : hdrs
       -- building with ShowS
       $ 'b' : 'y': 't' : 'e' : 's' : ' '
       : (if beg > end then ('*':) else
-          (showInt beg)
+          showInt beg
           . ('-' :)
-          . (showInt end))
+          . showInt end)
       ( '/'
       : showInt total "")
 
@@ -425,6 +449,6 @@ replaceHeader k v hdrs = (k,v) : deleteBy ((==) `on` fst) (k,v) hdrs
 
 composeHeaderBuilder :: H.HttpVersion -> H.Status -> H.ResponseHeaders -> Bool -> IO Builder
 composeHeaderBuilder ver s hs True =
-    fromByteString <$> composeHeader ver s (addTransferEncoding hs)
+    byteString <$> composeHeader ver s (addTransferEncoding hs)
 composeHeaderBuilder ver s hs False =
-    fromByteString <$> composeHeader ver s hs
+    byteString <$> composeHeader ver s hs
