@@ -25,6 +25,7 @@ import Network.Wai.Handler.Warp.Buffer
 import Network.Wai.Handler.Warp.Counter
 import qualified Network.Wai.Handler.Warp.Date as D
 import qualified Network.Wai.Handler.Warp.FdCache as F
+import Network.Wai.Handler.Warp.HTTP2
 import Network.Wai.Handler.Warp.Header
 import Network.Wai.Handler.Warp.ReadInt
 import Network.Wai.Handler.Warp.Recv
@@ -45,21 +46,21 @@ import System.Posix.IO (FdOption(CloseOnExec), setFdOption)
 import Network.Socket (fdSocket)
 #endif
 
--- | Default action value for 'Connection'.
+-- | Creating 'Connection' for plain HTTP based on a given socket.
 socketConnection :: Socket -> IO Connection
 socketConnection s = do
-    readBuf <- allocateBuffer bufferSize
+    bufferPool <- newBufferPool
     writeBuf <- allocateBuffer bufferSize
+    let sendall = Sock.sendAll s
     return Connection {
         connSendMany = Sock.sendMany s
-      , connSendAll = Sock.sendAll s
-      , connSendFile = defaultSendFile s
-      , connClose = sClose s >> freeBuffer readBuf >> freeBuffer writeBuf
-      , connRecv = receive s readBuf bufferSize
-      , connReadBuffer = readBuf
+      , connSendAll = sendall
+      , connSendFile = sendFile s writeBuf bufferSize sendall
+      , connClose = sClose s >> freeBuffer writeBuf
+      , connRecv = receive s bufferPool
+      , connRecvBuf = receiveBuf s
       , connWriteBuffer = writeBuf
       , connBufferSize = bufferSize
-      , connSendFileOverride = Override s
       }
 
 #if __GLASGOW_HASKELL__ < 702
@@ -67,14 +68,14 @@ allowInterrupt :: IO ()
 allowInterrupt = unblock $ return ()
 #endif
 
--- | Run an 'Application' on the given port. This calls 'runSettings' with
--- 'defaultSettings'.
+-- | Run an 'Application' on the given port.
+-- This calls 'runSettings' with 'defaultSettings'.
 run :: Port -> Application -> IO ()
 run p = runSettings defaultSettings { settingsPort = p }
 
--- | Run an 'Application' on the port present in the @PORT@ environment variable
---
--- Uses the 'Port' given when the variable is unset.
+-- | Run an 'Application' on the port present in the @PORT@
+-- environment variable. Uses the 'Port' given when the variable is unset.
+-- This calls 'runSettings' with 'defaultSettings'.
 --
 -- Since 3.0.9
 runEnv :: Port -> Application -> IO ()
@@ -90,6 +91,8 @@ runEnv p app = do
         _ -> fail $ "Invalid value in $PORT: " ++ sp
 
 -- | Run an 'Application' with the given 'Settings'.
+-- This opens a listen socket on the port defined in 'Settings' and
+-- calls 'runSettingsSocket'.
 runSettings :: Settings -> Application -> IO ()
 runSettings set app = withSocketsDo $
     bracket
@@ -99,15 +102,17 @@ runSettings set app = withSocketsDo $
             setSocketCloseOnExec socket
             runSettingsSocket set socket app)
 
--- | Same as 'runSettings', but uses a user-supplied socket instead of opening
--- one. This allows the user to provide, for example, Unix named socket, which
+-- | This installs a shutdown handler for the given socket and
+-- calls 'runSettingsConnection' with the default connection setup action
+-- which handles plain (non-cipher) HTTP.
+-- When the listen socket in the second argument is closed, all live
+-- connections are gracefully shut down.
+--
+-- The supplied socket can be a Unix named socket, which
 -- can be used when reverse HTTP proxying into your application.
 --
 -- Note that the 'settingsPort' will still be passed to 'Application's via the
 -- 'serverPort' record.
---
--- When the listen socket in the second argument is closed, all live
--- connections are gracefully shut down.
 runSettingsSocket :: Settings -> Socket -> Application -> IO ()
 runSettingsSocket set socket app = do
     settingsInstallShutdownHandler set closeListenSocket
@@ -125,10 +130,13 @@ runSettingsSocket set socket app = do
 
     closeListenSocket = sClose socket
 
--- | Allows you to provide a function which will return a 'Connection'. In
--- cases where creating the @Connection@ can be expensive, this allows the
--- expensive computations to be performed in a separate thread instead of the
--- main server loop.
+-- | The connection setup action would be expensive. A good example
+-- is initialization of TLS.
+-- So, this converts the connection setup action to the connection maker
+-- which will be executed after forking a new worker thread.
+-- Then this calls 'runSettingsConnectionMaker' with the connection maker.
+-- This allows the expensive computations to be performed
+-- in a separate worker thread instead of the main server loop.
 --
 -- Since 1.3.5
 runSettingsConnection :: Settings -> IO (Connection, SockAddr) -> Application -> IO ()
@@ -138,6 +146,8 @@ runSettingsConnection set getConn app = runSettingsConnectionMaker set getConnMa
       (conn, sa) <- getConn
       return (return conn, sa)
 
+-- | This modifies the connection maker so that it returns 'TCP' for 'Transport'
+-- (i.e. plain HTTP) then calls 'runSettingsConnectionMakerSecure'.
 runSettingsConnectionMaker :: Settings -> IO (IO Connection, SockAddr) -> Application -> IO ()
 runSettingsConnectionMaker x y =
     runSettingsConnectionMakerSecure x (go y)
@@ -146,8 +156,10 @@ runSettingsConnectionMaker x y =
 
 ----------------------------------------------------------------
 
--- | Allows you to provide a function which will return a function
--- which will return 'Connection'.
+-- | The core run function which takes 'Settings',
+-- a connection maker and 'Application'.
+-- The connection maker can return a connection of either plain HTTP
+-- or HTTP over TLS.
 --
 -- Since 2.1.4
 runSettingsConnectionMakerSecure :: Settings -> IO (IO (Connection, Transport), SockAddr) -> Application -> IO ()
@@ -168,11 +180,6 @@ runSettingsConnectionMakerSecure set getConnMaker app = do
                    T.stopManager
                    f
 
-onE :: Settings -> Maybe Request -> SomeException -> IO ()
-onE set mreq e = case fromException e of
-    Just (NotEnoughLines []) -> return ()
-    _                        -> settingsOnException set mreq e
-
 -- Note that there is a thorough discussion of the exception safety of the
 -- following code at: https://github.com/yesodweb/wai/issues/146
 --
@@ -198,7 +205,7 @@ acceptConnection set getConnMaker app dc fc tm counter = do
     -- First mask all exceptions in acceptLoop. This is necessary to
     -- ensure that no async exception is throw between the call to
     -- acceptNewConnection and the registering of connClose.
-    void $ mask_ $ acceptLoop
+    void $ mask_ acceptLoop
     gracefulShutdown counter
   where
     acceptLoop = do
@@ -224,7 +231,7 @@ acceptConnection set getConnMaker app dc fc tm counter = do
         case ex of
             Right x -> return $ Just x
             Left  e  -> do
-                onE set Nothing $ toException e
+                settingsOnException set Nothing $ toException e
                 if isFullErrorType (ioeGetErrorType e) then do
                     -- "resource exhausted (Too many open files)" may
                     -- happen by accept().  Wait a second hoping that
@@ -257,21 +264,20 @@ fork set mkConn addr app dc fc tm counter = settingsFork set $ \ unmask ->
     -- We grab the connection before registering timeouts since the
     -- timeouts will be useless during connection creation, due to the
     -- fact that async exceptions are still masked.
-    bracket mkConn closeConn $ \(conn0, transport) ->
+    bracket mkConn closeConn $ \(conn, transport) ->
 
     -- We need to register a timeout handler for this thread, and
     -- cancel that handler as soon as we exit.
     bracket (T.registerKillThread tm) T.cancel $ \th ->
 
-    let ii = InternalInfo th fc dc
-        conn = setSendFile conn0 fc
+    let ii = InternalInfo th tm fc dc
         -- We now have fully registered a connection close handler
         -- in the case of all exceptions, so it is safe to one
         -- again allow async exceptions.
     in unmask .
        -- Call the user-supplied on exception code if any
        -- exceptions are thrown.
-       handle (onE set Nothing) .
+       handle (settingsOnException set Nothing) .
 
        -- Call the user-supplied code for connection open and close events
        bracket (onOpen addr) (onClose addr) $ \goingon ->
@@ -293,13 +299,28 @@ serveConnection :: Connection
                 -> Application
                 -> IO ()
 serveConnection conn ii origAddr transport settings app = do
-    istatus <- newIORef False
-    src <- mkSource (connSource conn th istatus)
-    addr <- getProxyProtocolAddr src
-    http1 addr istatus src `E.catch` \e -> do
-        sendErrorResponse addr istatus e
-        throwIO (e :: SomeException)
-
+    -- fixme: Upgrading to HTTP/2 should be supported.
+    (h2,bs) <- if isHTTP2 transport then
+                   return (True, "")
+                 else do
+                   bs0 <- connRecv conn
+                   if S.length bs0 >= 4 && "PRI " `S.isPrefixOf` bs0 then
+                       return (True, bs0)
+                     else
+                       return (False, bs0)
+    if h2 then do
+        recvN <- makeReceiveN bs (connRecv conn) (connRecvBuf conn)
+        -- fixme: origAddr
+        http2 conn ii origAddr transport settings recvN app
+      else do
+        istatus <- newIORef False
+        src <- mkSource (wrappedRecv conn th istatus (settingsSlowlorisSize settings))
+        writeIORef istatus True
+        leftoverSource src bs
+        addr <- getProxyProtocolAddr src
+        http1 addr istatus src `E.catch` \e -> do
+            sendErrorResponse addr istatus e
+            throwIO (e :: SomeException)
   where
     getProxyProtocolAddr src =
         case settingsProxyProtocol settings of
@@ -361,8 +382,8 @@ serveConnection conn ii origAddr transport settings app = do
             `E.catch` \e -> do
                 -- Call the user-supplied exception handlers, passing the request.
                 sendErrorResponse addr istatus e
-                onE settings (Just req) e
-                -- Don't throw the error again to prevent calling onE twice.
+                settingsOnException settings (Just req) e
+                -- Don't throw the error again to prevent calling settingsOnException twice.
                 return False
         when keepAlive $ http1 addr istatus src
 
@@ -390,8 +411,8 @@ serveConnection conn ii origAddr transport settings app = do
         -- We just send a Response and it takes a time to
         -- receive a Request again. If we immediately call recv,
         -- it is likely to fail and the IO manager works.
-        -- It is very costy. So, we yield to another Haskell
-        -- thread hoping that the next Request will arraive
+        -- It is very costly. So, we yield to another Haskell
+        -- thread hoping that the next Request will arrive
         -- when this Haskell thread will be re-scheduled.
         -- This improves performance at least when
         -- the number of cores is small.
@@ -399,7 +420,7 @@ serveConnection conn ii origAddr transport settings app = do
 
         if not keepAlive then
             return False
-          else do
+          else
             -- If there is an unknown or large amount of data to still be read
             -- from the request body, simple drop this connection instead of
             -- reading it all in to satisfy a keep-alive request.
@@ -449,12 +470,12 @@ flushBody src =
                 | toRead' >= 0 -> loop toRead'
                 | otherwise -> return False
 
-connSource :: Connection -> T.Handle -> IORef Bool -> IO ByteString
-connSource Connection { connRecv = recv } th istatus = do
+wrappedRecv :: Connection -> T.Handle -> IORef Bool -> Int -> IO ByteString
+wrappedRecv Connection { connRecv = recv } th istatus slowlorisSize = do
     bs <- recv
     unless (S.null bs) $ do
         writeIORef istatus True
-        when (S.length bs >= 2048) $ T.tickle th
+        when (S.length bs >= slowlorisSize) $ T.tickle th
     return bs
 
 -- Copied from: https://github.com/mzero/plush/blob/master/src/Plush/Server/Warp.hs
@@ -467,8 +488,4 @@ setSocketCloseOnExec socket =
 #endif
 
 gracefulShutdown :: Counter -> IO ()
-gracefulShutdown counter = do
-    -- To avoid race condition, we just use threadDelay, not MVar.
-    threadDelay 10000000
-    noConnections <- isZero counter
-    unless noConnections $ gracefulShutdown counter
+gracefulShutdown counter = waitForZero counter
