@@ -9,14 +9,17 @@ import Data.ByteString (ByteString)
 import qualified Data.ByteString as S
 import Data.IORef (IORef, readIORef, writeIORef, newIORef)
 import Data.Typeable (Typeable)
-import Data.Word (Word16)
+import Data.Word (Word16, Word8)
+import Foreign.Ptr (Ptr)
 import Network.HTTP.Types.Header
-import Network.Socket (Socket)
-import Network.Wai.Handler.Warp.Buffer (Buffer,BufSize)
 import qualified Network.Wai.Handler.Warp.Date as D
 import qualified Network.Wai.Handler.Warp.FdCache as F
 import qualified Network.Wai.Handler.Warp.Timeout as T
 
+#ifndef WINDOWS
+import System.Posix.Types (Fd)
+#endif
+
 ----------------------------------------------------------------
 
 -- | TCP port number.
@@ -64,25 +67,61 @@ instance Exception InvalidRequest
 
 ----------------------------------------------------------------
 
--- | Whether or not 'ConnSendFileOverride' in 'Connection' can be
---   overridden. This is a kind of hack to keep the signature of
---   'Connection' clean.
-data ConnSendFileOverride = NotOverride     -- ^ Don't override
-                          | Override Socket -- ^ Override with this 'Socket'
+#ifdef WINDOWS
+type Fd = ()
+#endif
 
-----------------------------------------------------------------
+-- | Data type to abstract file identifiers.
+--   On Unix, a file descriptor would be specified to make use of
+--   the file descriptor cache.
+--
+-- Since: 3.1.0
+data FileId = FileId {
+    fileIdPath :: FilePath
+  , fileIdFd   :: Maybe Fd
+  }
+
+-- |  fileid, offset, length, hook action, HTTP headers
+--
+-- Since: 3.1.0
+type SendFile = FileId -> Integer -> Integer -> IO () -> [ByteString] -> IO ()
+
+-- | Type for read buffer pool
+type BufferPool = IORef ByteString
+
+-- | Type for buffer
+type Buffer = Ptr Word8
+
+-- | Type for buffer size
+type BufSize = Int
+
+-- | Type for the action to receive input data
+type Recv = IO ByteString
+
+-- | Type for the action to receive input data with a buffer.
+--   The result boolean indicates whether or not the buffer is fully filled.
+type RecvBuf = Buffer -> BufSize -> IO Bool
 
 -- | Data type to manipulate IO actions for connections.
-data Connection = Connection
-    { connSendMany :: [ByteString] -> IO ()
-    , connSendAll  :: ByteString -> IO ()
-    , connSendFile :: FilePath -> Integer -> Integer -> IO () -> [ByteString] -> IO () -- ^ filepath, offset, length, hook action, HTTP headers
-    , connClose    :: IO ()
-    , connRecv     :: IO ByteString
-    , connReadBuffer       :: Buffer
-    , connWriteBuffer      :: Buffer
-    , connBufferSize       :: BufSize
-    , connSendFileOverride :: ConnSendFileOverride
+--   This is used to abstract IO actions for plain HTTP and HTTP over TLS.
+data Connection = Connection {
+    -- | This is not used at this moment.
+      connSendMany    :: [ByteString] -> IO ()
+    -- | The sending function.
+    , connSendAll     :: ByteString -> IO ()
+    -- | The sending function for files in HTTP/1.1.
+    , connSendFile    :: SendFile
+    -- | The connection closing function.
+    , connClose       :: IO ()
+    -- | The connection receiving function. This returns "" for EOF.
+    , connRecv        :: Recv
+    -- | The connection receiving function. This tries to fill the buffer.
+    --   This returns when the buffer is filled or reaches EOF.
+    , connRecvBuf     :: RecvBuf
+    -- | The write buffer.
+    , connWriteBuffer :: Buffer
+    -- | The size of the write buffer.
+    , connBufferSize  :: BufSize
     }
 
 ----------------------------------------------------------------
@@ -90,12 +129,14 @@ data Connection = Connection
 -- | Internal information.
 data InternalInfo = InternalInfo {
     threadHandle :: T.Handle
+  , timeoutManager :: T.Manager
   , fdCacher :: Maybe F.MutableFdCache
   , dateCacher :: D.DateCache
   }
 
 ----------------------------------------------------------------
 
+-- | Type for input streaming.
 data Source = Source !(IORef ByteString) !(IO ByteString)
 
 mkSource :: IO ByteString -> IO Source
